#!/bin/bash
#
# Copyright (c) 2014 Joe Linoff
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# ================================================================
# Functions
# ================================================================
function help() {
    bn=$(basename $0)
    cat <<EOF

USAGE
  $bn [OPTIONS]

DESCRIPTION
  This script will install a web site in a completely self contained
  directory tree on this host that does not require administration
  privileges to run. It also creates a customized script that you
  can use later for other installations.

  You can remove this installation at any time by simply stopping the
  site using the <site>/rtf/bin/sitemgr.py tool.

  There are number of command line options available. They are
  described below. If you do not specify them then you will be
  prompted for them with defaults.

OPTIONS
  -a <name>, -app <name>   The app name.
                           The default is "myapp".

  -c <file>, --cfg <file>  The cfg file. This is where the install
                           options are read from and save to.
                           The default is install.cfg.

  --db-port <port>         The database port to use.
                           The default is 9082.

  -h, --help               This help message.

  --http-port <port>       The HTTP port to use.
                           The default is 9080.

  --https-port <port>      The HTTPS port to use.
                           The default is 9081.

  -p, --prompt             Force all prompts.
                           This is useful when a configuration
                           file is present.

  --ports <HTTP> <HTTPS> <DB>
                           Specify all of the ports at once separated
                           by commas in order HTTP, HTTPS, DB.

  -r <dir>, --root <dir>   The root directory for the site.
                           The default is the current directory.

  -s <name>, --site <name> The site name. The full path to the
                           site will be: <root>/<site>.
                           The default is "mywebsite".

  --start                  Start the site automatically after
                           installation.

  -v, --verbose            Increase the level of verbosity.

EXAMPLES
  1. Run with prompts:
     \$ $0

  2. Dry run with prompts:
     \$ $0 -n

  3. Run with the configuration file, force prompts:
     \$ # Assumes that a previous run was done.
     \$ $0 -p

  4. Run with some command line options:
     \$ $0 --ports 10080 10081 10082 -a introduction

EOF
}

function vinfo() {
    if (( $1 )) ; then
        echo $2
    fi
}

function err() {
    local msg="$1"
    cat <<EOF

ERROR: $msg

EOF
    exit 1
}

function prompt() {
    local text="$1"
    local default="$2"
    local p="$text? "
    if [[ "$default" != "" ]] ; then
        p="$text <$default>? "
    fi
    local val=""
    read -p "$p" val
    if [[ "$val" == "" ]] ; then
        val="$default"
    fi
    echo $val
}

function check_next_arg() {
    local num=$1
    local opt=$2
    if (( $num == 0 )) ; then
        err "missing argument for $opt"
    fi
}

function check_num() {
    local num="$1"
    local opt="$2"
    local re='^[0-9]+'
    if ! [[ "$num" =~ $re ]] ; then
        err "expected a number for '$opt' but found '$num'"
    fi
}

function abspath() {
    local relpath="$1"
    python -c "import os; print os.path.abspath('$relpath')"
}

# ================================================================
# Options
# ================================================================
app_name=""
db_port=0
dryrun=""
http_port=0
https_port=0
prompt=0
root_dir=""
site_name=""
start=0
cfg="install.cfg"
v=0

if [ -f "$cfg" ] ; then
    # If a configuration file is found, use it.
    . $cfg
fi

while (( $# )) ; do
    opt=$1
    shift
    case $opt in
        -a|--app)
            check_next_arg $# $opt
            app_name="$1"
            shift
            ;;
	-c|--cfg)
            check_next_arg $# $opt
            cfg="$1"
            shift
            ;;
        --db-port)
            check_num "$1" $opt
            db_port=$1
            shift
            ;;
        -h|--help)
            help
            exit 0
            ;;
        --http-port)
            check_num "$1" $opt
            http_port=$1
            shift
            ;;
        --https-port)
            check_num "$1" $opt
            https_port=$1
            shift
            ;;
	-n)
	    dryrun="echo"
	    ;;
	-p|--prompt)
	    prompt=1
	    ;;
	--ports)
            check_num "$1" "$opt <HTTP port>"
            check_num "$2" "$opt <HTTPS port>"
            check_num "$3" "$opt <DB port>"

	    http_port=$1
	    https_port=$2
	    db_port=$3

	    shift
	    shift
	    shift
	    ;;
        -r|--root)
            check_next_arg $# $opt
            root_dir="$1"
            shift
            ;;
        -s|--site)
            check_next_arg $# $opt
            site_name="$1"
            shift
            ;;
        --start)
            start=1
            ;;
        -v|--verbose)
            v=$(( $v + 1 ))
            ;;
        *)
            err "unrecognized option '$opt'."
            ;;
    esac
done


# ================================================================
# Environment check
# ================================================================
vinfo $v "checking the environment"
required_programs=(python make gcc openssl)
for prog in ${required_programs[@]} ; do
    vinfo $v "checking $prog"
    which $prog &>/dev/null
    st=$?
    if (( $st )) ; then
        err "required program not found: $prog"
    fi
done

# Linux/Mac way to get the source directory.
medir=$(dirname $(echo $(cd $(dirname "$0"); pwd)/$(basename "$0")))
installer=$medir/mkwebsite.py
vinfo $v "checking the for the existence of the installer: $installer"
if [ ! -f $installer ] ; then
    err "cannot find $installer"
fi
version=$(grep ^VERSION $medir/mkwebsite.py | awk -F= '{print $2;}' | awk -F\' '{print $2;}')


# ================================================================
# help
# ================================================================
cat <<EOF
# ****************************************************************
# This is the installer for mkwebsite-$version.
#
# Would you like to see the help before proceeding?
# 
# Please enter yes or no.
EOF
answer=$(prompt "Help" "yes")
if [[ "$answer" == "yes" ]] ; then
    help
fi


# ================================================================
# Prompt for the root directory.
# ================================================================
if [[ "$root_dir" == "" ]] || (( $prompt )) ; then
    cat <<EOF
# ****************************************************************
# Please enter the root directory for the installation.
# 
# The default is the current directory: $(pwd)
# All relative paths are relative to it.
EOF
    if [[ "$root_dir" == "" ]] ; then
	root_dir=$(pwd)
    fi
    root_dir=$(prompt "Root directory" "$root_dir")
    root_dir=$(abspath $root_dir)
fi


# ================================================================
# Prompt for the site directory.
# ================================================================
if [[ "$site_name" == "" ]] || (( $prompt )) ; then
    cat <<EOF

# ****************************************************************
# Please enter the site name.
# 
# This is a subdirectory under the root directory that contains
# this site. With this approach you can have multiple sites in
# the same directory tree, for example:
# 
#    /opt/websites/production
#    /opt/websites/development
# 
# Where /opt/websites is the root directory with two sites:
# production and development.
EOF
    if [[ "$site_name" == "" ]] ; then
	site_name="mywebsite"
    fi
    site_name=$(prompt "Site name" $site_name)
fi


# ================================================================
# Prompt for the app name.
# ================================================================
if [[ "$app_name" == "" ]] || (( $prompt )) ; then
    cat <<EOF

# ****************************************************************
# Please enter your app name.
# 
# This is the name of the initial app name that is set up for you in the
# Django directory structure. You can change it later, you can delete it
# and you can additional apps so the name isn't that important.
EOF
    if [[ "$app_name" == "" ]] ; then
	app_name="myapp"
    fi
    app_name=$(prompt "App name" $app_name)
fi


# ================================================================
# Prompt for the HTTP port.
# ================================================================
if (( $http_port == 0 )) || (( $prompt )) ; then
    cat <<EOF

# ****************************************************************
# Please enter the HTTP port value.
# 
# This is the port that you use to access the website in your
# browser. If you want to use port 80 you will have to either
# install the website as root (bad idea for security reasons)
# or update your firewall to forward traffic between port 80
# and this port.
EOF
    if (( $http_port == 0 )) ; then
	http_port=9080
    fi
    port=$(prompt "HTTP port" $http_port)
    re='^[0-9]+'
    if ! [[ "$port" =~ $re ]] ; then
        err "expected a number for the HTTP port but found '$port'"
    fi
    if (( $port == $https_port )) ; then
        err "HTTP port must be different than the HTTPS port"
    fi
    if (( $port == $db_port )) ; then
        err "HTTP port must be different than the HTTPS port"
    fi
    http_port=$port
fi


# ================================================================
# Prompt for the HTTPS port.
# ================================================================
if (( $https_port == 0 )) || (( $prompt )) ; then
    cat <<EOF

# ****************************************************************
# Please enter the HTTPS port value.
# 
# This is the port that you use to access the website in your
# browser. If you want to use port 443 you will have to either
# install the website as root (bad idea for security reasons)
# or update your firewall to forward traffic between port 443
# and this port.
EOF
    if (( $https_port == 0 )) ; then
	https_port=9081
    fi
    port=$(prompt "HTTPS port" $https_port)
    re='^[0-9]+'
    if ! [[ "$port" =~ $re ]] ; then
	err "expected a number for the HTTPS port but found '$port'"
    fi
    if (( $port == $http_port )) ; then
        err "HTTPS port must be different than the HTTP port"
    fi
    if (( $port == $db_port )) ; then
        err "HTTPS port must be different than the database port"
    fi
    https_port=$port
fi


# ================================================================
# Prompt for the database port.
# ================================================================
if (( $db_port == 0 )) || (( $prompt )) ; then
    cat <<EOF

# ****************************************************************
# Please enter the database port value.
# 
# This is the port that you use to access your database server.
EOF
    if (( $db_port == 0 )) ; then
	db_port=9082
    fi
    port=$(prompt "Database port" $db_port)
    re='^[0-9]+'
    if ! [[ "$port" =~ $re ]] ; then
        err "expected a number for the HTTP port but found '$port'"
    fi
    if (( $port == $http_port )) ; then
        err "database port must be different than the HTTP port"
    fi
    if (( $port == $https_port )) ; then
        err "database port must be different than the HTTPS port"
    fi
    db_port=$port
fi


# ================================================================
# Options summary.
# ================================================================
start_token="no"
if (( $start )) ; then
    start_token="yes"
fi
cat <<EOF

Configuration Options:

    App Name      : "$app_name"
    Configuration : "$cfg"
    Database Port : $db_port
    HTTP Port     : $http_port
    HTTPS Port    : $https_port
    Log File      : "install.log"
    Root Dir      : "$root_dir"
    Site Name     : "$site_name"
    Start         : $start_token
    Website Dir   : "$root_dir/$site_name"
    Verbosity     : $v
    Version       : $version

EOF

    cat <<EOF

# ****************************************************************
# Are the above configuration options correct?
# 
# Please enter yes if they are or no if they are not.
EOF
continue=$(prompt "Correct" "yes")
if ! [[ "$continue" == "yes" ]] ; then
    echo "Please rerun with the -p option to fix problems."
    exit 0
fi


vinfo $v "creating $cfg"
cat >$cfg <<EOF
# Configuration options created `date`.
app_name="$app_name"
db_port=$db_port
http_port=$http_port
https_port=$https_port
root_dir="$root_dir"
site_name="$site_name"
EOF


# ================================================================
# Run the real installer.
# ================================================================
optsws=""
if [ -d $medir/downloads ] ; then
    optsws="-d $medir/downloads "
fi
if (( $start )) ; then
    optsws="$optsws --start "
fi

vinfo $v "running the actual installer"
$dryrun time $installer -v \
    $optsws \
    -r $root_dir \
    -s $site_name \
    -a $app_name \
    --port-http $http_port \
    --port-https $https_port \
    --port-db $db_port \
    2>&1 | tee install.log

# ================================================================
# Save the configuration in the website directory after successful
# completion.
# ================================================================
vinfo $v "copying the configuration to $root_dir/$site_name"
cp $cfg $root_dir/$site_name/

