#!/usr/bin/env python
r'''
Setting up a web site is a complex problem because there are a lot of
intersecting parts. This tool creates a complete website using django,
postgresql, nginx, uwsgi, memcached and various static technologies
like jquery, jquery-ui, tooltipster, datatables, dynatree, bootstrap
and cryptojs in a directory that you specify (even your home
directory) that does not require root permissions to run. It also
installs a number of useful python packages that you can use in your
django apps like fabric, south and yolk. It is ideal for learning
about website technology, experimenting with features and for
developing complete websites.

It also creates a tool for starting, stopping, restarting and checking
the website as well a sample page that can be used as a starting
point.

The goal is to allow you to get a production website up and running
in 30 minutes or less on non-privileged ports.

Since the site uses python (i.e. django) this script is written in
python for maximum portability. You will need python 2.7 unless you
install specific packages like argparse.

You will need gcc, make and other development tools installed.

To use it simply type:

    $ python mkwebsite.py -v \
         -r /opt/websites \
         -d /tmp/downloads \
         -s mysite \
         -a myapp

This will create the installation in /opt/websites/mysite.

The -d switch archives the download files for re-use.

The -r switch specifies the root directory.

The -s switch specifies the django project name (the sitename).

The -a switch specifies the django app name. The appname must be
different than the sitename.

When the bootstrap process completes it will print instructions that
describe how to start the website services and access it. For the
example above you would start the services like this:

    $ /opt/websites/mysite/rtf/bin/mysite-mgr.py start

There are additional options that give you fine grain control over
things like the port values used for the different services. You
will need to use them if you want to have multiple site installations
on the same server to avoid port conflicts.

Note that I considered adding additional python packages like twisted and
boto to the virtualenv but decided against it in the end because they
could be considered more special purpose. If you wish to install them
or any other package, simply setup the virtualenv and run pip. For
the example above where the sitename is "mysite" in /opt/websitse you
would do this:

    $ . /opt/websites/mysite/venv/bin/activate
    (venv) $ pip install boto twisted

I have tested it on CentOS 5.x linux, CentOS 6.x linux and Mac OS X
10.9.1.'''
import argparse
import datetime
import getpass
import grp
import inspect
import logging
import os
import pwd
import shutil
import socket
import stat
import sys
import subprocess
import tarfile
import time
import urllib2
import zipfile
from pprint import pprint

VERSION = '0.16'


# ================================================================
# __total_seconds
# ================================================================
def _total_seconds(elapsed):
    '''
    Return the total seconds elapsed if total_seconds() is not
    available.
    '''
    if 'total_seconds' in dir(datetime.timedelta):
        secs = elapsed.total_seconds()  # use the native version if available
    else:
        usecs = float(elapsed.microseconds)
        secs = (usecs + (elapsed.seconds + elapsed.days * 24. * 3600.) * 10**6) / 10**6
    return secs


# ================================================================
# _runcmd
# ================================================================
def _runcmd(cmd):
    '''
    Execute a shell command, return the output as a string as well as
    the status.
    '''
    # Would love to use subprocess.check_output() but doesn't
    # work for older versions of python. I can emulate it
    # but the subprocess.CalledProcessError exception also changed.
    # In earlier version it did not have the output attribute.
    # This approach is just easier.
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
    out, _ = process.communicate()
    sts = process.poll()
    return sts, out


# ================================================================
# runcmd
# ================================================================
def runcmd(context, cmd, exit_on_error=True):
    opts = context['opts']
    logger = context['logger']
    start = datetime.datetime.now()
    lineno = inspect.stack()[1][2]  # callers line number

    logger.info('cmd-begin:%d: %s' % (lineno, cmd))
    sts, out = _runcmd(cmd)

    if opts.verbose:
        sys.stdout.write(out)
        sys.stdout.flush()

    elapsed = datetime.datetime.now() - start
    logger.info('cmd-end:%d: %.1f seconds' % (lineno, _total_seconds(elapsed)))

    if sts != 0:
        logger.error('command failed with status %d: %s' % (sts, cmd))
        if exit_on_error is True:
            sys.exit(1)
    return sts, out


# ================================================================
# makedirs
# ================================================================
def makedirs(context, path, mode=0777):
    '''
    Make the directory if it does not already exist.
    '''
    if os.path.exists(path):
        return

    logger = context['logger']
    logger.info('making directory "%s"' % (path))
    try:
        os.makedirs(path, mode)
    except OSError as exc:
        logger.error(repr(exc))
        if exc[0] == 13:
            logger.error('insufficient permissions to create directory: ' + path)
        sys.exit(1)


# ================================================================
# walkdirs
# ================================================================
def walkdirs(root, norecurse=['pkg',]):
    '''
    Walk a directory tree and capture the results.
    '''
    dirtree = []
    def _walkdirs(root, dirname, level=4, recurse=True):
        indent = (level*2)*' '
        rootdir = os.path.join(root, dirname)

        entries = sorted(os.listdir(rootdir), key=str.lower)
        for entry in entries:
            path = os.path.join(rootdir, entry)
            if os.path.isdir(path):
                if recurse is True:
                    dirtree.append(indent + '<li>' + entry)
                    indent1 = ((level+1)*2)*' '
                    dirtree.append(indent1 + '<ul>')
                    flag = entry not in norecurse
                    _walkdirs(rootdir, entry, level+2, flag)
                    dirtree.append(indent1 + '</ul>')
                    dirtree.append(indent + '</li> <!-- ' + entry + ' -->')
                else:
                    dirtree.append(indent + '<li>' + entry + ' ...' + '</li>')

        for entry in entries:
            path = os.path.join(rootdir, entry)
            if os.path.isdir(path) is False:
                dirtree.append(indent + '<li>' + entry + '</li>')
    _walkdirs('', root)
    return dirtree


# ================================================================
# port_is_open
# ================================================================
def port_is_open(address, port):
    '''
    Check to see if a port is open.
    '''
    try:
        socket.socket().connect((address, port))
        return True
    except socket.error:
        return False


# ================================================================
# backup
# ================================================================
def backup(context, fname):
    '''
    Backup the existing file.
    '''
    logger = context['logger']
    for i in range(1, 1000):
        ofn = '%s-%04d' % (fname, i)
        if os.path.exists(ofn) is False:
            logger.info('backing up to ' + ofn)
            shutil.copyfile(fname, ofn)
            return


# ================================================================
# init_logger
# ================================================================
def init_logger(name, level=logging.INFO):
    '''
    Initialize the logging.
    '''
    logger = logging.getLogger(name)
    lch = logging.StreamHandler()

    fmt = '%(asctime)s %(levelname)7s %(filename)s:%(lineno)05d %(message)s'
    formatter = logging.Formatter(fmt)

    lch.setFormatter(formatter)
    logger.addHandler(lch)

    logger.setLevel(level)
    return logger


# ================================================================
# wget
# ================================================================
def wget(logger, url, ofn):
    start = datetime.datetime.now()
    lineno = inspect.stack()[1][2]  # callers line number
    logger.info('wget:%d %s' % (lineno, url))
    logger.info('  to:%d %s' % (lineno, ofn))
    hdrs = {'Accept': '*/*',
            'Connection': 'keep-alive',
            'User-Agent': 'Wget/1.14',
        }
    request = urllib2.Request(url, headers=hdrs)
    page = urllib2.urlopen(request, timeout=10)
    contents = page.read()

    ofp = open(ofn, 'wb')
    ofp.write(contents)
    ofp.close()

    elapsed = datetime.datetime.now() - start
    logger.info('downloaded %d bytes in %.1f seconds' % (len(contents),
                                                         _total_seconds(elapsed)))


# ================================================================
# pip_install
# ===========================================<=====================
def pip_install(context, pkg, exit_on_error=True):
    '''
    Pip install a python package.
    '''
    # Determine whether the package has already been installed.
    if pkg in context['venv']['installed']:  # O(N) search but N is very small.
        return  # already installed

    # See if the package was install manually by
    # looking for a rough name match.
    pkgdir = context['venv']['site-packages']
    for item in os.listdir(pkgdir):
        item = item.lower()
        lpkg = pkg.lower()
        if item.find(lpkg) == 0:
            context['venv']['installed'].append(pkg)
            return

    # This package needs to be installed, tell the user.
    context['logger'].info('pip install ' + pkg)

    # Update the path.
    path = ''
    for xdir in context['dirs']:
        if xdir.find('/bin') > 0:
            path += context['dirs'][xdir] + ':'

    # pip install
    cmd = '%s && export PATH="%s:${PATH}" && pip install %s' % (context['venv_activate_cmd'], path, pkg)
    sts, out = runcmd(context, cmd, exit_on_error)
    if sts == 0:
        context['venv']['installed'].append(pkg)

    return sts, out
        

# ================================================================
# fixfn
# ================================================================
def fixfn(data):
    '''
    Fix the file name so that spaces and other odd characters are
    removed.
    '''
    return data.replace(' ','-').replace('%20','-')


# ================================================================
# download_and_extract
# ================================================================
def download_and_extract(context, url):
    '''
    Download and expand.
    '''
    logger = context['logger']
    basename = os.path.basename(url)
    ext = None
    bnm = None

    if basename[-8:] == '.tar.bz2':
        ext = basename[-8:]
        bnm = basename[:-8]
        mode = 'r:bz2'
    elif basename[-7:] == '.tar.gz':
        ext = basename[-7:]
        bnm = basename[:-7]
        mode = 'r:gz'
    else:
        logger.error('unrecognized extension in ' + basename)

    ofn = os.path.join(context['dirs']['download'], basename)
    if os.path.exists(ofn) is False:
        wget(logger, url, ofn)

    srcpath = os.path.join(context['dirs']['src'], bnm)
    if os.path.exists(srcpath) is False:
        logger.info('extracting ' + srcpath)
        tfp = tarfile.open(ofn, mode)
        tfp.extractall(context['dirs']['src'])
        tfp.close()

    return srcpath


# ================================================================
# install_static
# ================================================================
def install_static(context):
    '''
    Install the static web tools.
    '''
    logger = context['logger']
    pkgs = context['static_pkgs']

    # Download
    for pkg in pkgs:
        url = pkg[0]
        dfn = pkg[1]
        if os.path.exists(dfn) is False:
            wget(logger, url, dfn)

    # Extract
    for pkg in pkgs:
        dfn = pkg[1]
        pos = dfn.rfind('.')
        ext = dfn[pos:]
        if ext in ['.js', '.map']:
            dst = fixfn(os.path.join(context['dirs']['static/js'], os.path.basename(dfn)))
            if os.path.exists(dst) is False:
                logger.info('copy %s' % (dfn))
                logger.info('  to %s' % (dst))
                shutil.copyfile(dfn, dst)
        elif ext == '.zip':
            dstname = os.path.basename(dfn)[:-len(ext)]
            dst = fixfn(os.path.join(context['dirs']['static/pkg'], dstname))
            if os.path.exists(dst) is False:
                logger.info('unzip to %s' % (dst))
                zfo = zipfile.ZipFile(dfn, 'r')
                zfo.extractall(dst)
        else:
            logger.error('unrecognized extension: "%s" for "%s"' % (ext, dfn))
            sys.exit(1)


# ================================================================
# install_db
# ================================================================
def install_db(context):
    '''
    Install the DB.
    '''
    logger = context['logger']
    url = context['db']['url']
    srcpath = download_and_extract(context, url)

    # Configure
    rtfdir = context['dirs']['db']
    config_status = os.path.join(srcpath, 'config.status')
    if os.path.exists(config_status) is False:
        logger.info('configuring ' + srcpath)
        cmd = 'cd %s && ./configure --prefix="%s" --exec-prefix="%s" --with-pgport=%d' % (
            srcpath,
            rtfdir,
            rtfdir,
            context['ports']['db'],
        )
        sts, out = runcmd(context, cmd, exit_on_error=False)
        if sts != 0:
            # attempt to recover
            # discovered this problem on a server
            logger.info('reconfigure without readline')
            cmd += ' --without-readline'
            runcmd(context, cmd)

    # Make, install, update the database.
    rtfbindir = os.path.join(rtfdir, 'bin')
    semaphore = os.path.join(rtfbindir, 'pg_ctl')
    if os.path.exists(semaphore) is False:
        logger.info('compiling the database')
        cmd = 'cd %s && make' % (srcpath)
        runcmd(context, cmd)

        logger.info('installing the database')
        cmd = 'cd %s && make install' % (srcpath)
        runcmd(context, cmd)

    # Install psycopg2 python package
    found = sum([1 for _, dirs, _ in os.walk(context['dirs']['venv/lib']) if 'psycopg2' in dirs])
    if found == 0:
        logger.info('installing psycopg2')

        # set the environment
        rtflibdir = os.path.join(rtfdir, 'lib')
        if 'LD_LIBRARY_PATH' in os.environ:
            os.environ['LD_LIBRARY_PATH'] += os.pathsep + rtflibdir
        else:
            os.environ['LD_LIBRARY_PATH'] = rtflibdir

        pip_install(context, 'psycopg2')

    dbdata = context['dirs']['dbdata']
    dbsem = os.path.join(dbdata, 'PG_VERSION')
    if os.path.exists(dbsem) is False:
        cmd = 'export PATH="%s:${PATH}" && pg_ctl -D %s initdb' % (rtfbindir, dbdata)
        logger.info('initializing the database as %s' % (context['username']))
        if context['login'] == 'root':
            # special case handling for root because the DB cannot be initialized as root.
            cmd = "su -m %s -c '%s'" % (context['username'], cmd)
            os.chown(dbdata, context['uid'], context['gid'])
            os.chmod(dbdata, 0777)
        runcmd(context, cmd)

    # At this point the server can be started using either of these commands.
    logfile = os.path.join(context['dirs']['log'], 'postgresql.log')
    context['db']['logfile'] = logfile

    # Used during Django configuration.
    dbdata = context['dirs']['dbdata']
    pgctl = os.path.join(rtfbindir, 'pg_ctl') + ' -D ' + dbdata
    context['db']['start'] = pgctl + ' -l ' + logfile + ' start'
    context['db']['stop'] = pgctl + ' stop'


# ================================================================
# install_pcre
# ================================================================
def install_pcre(context):
    '''
    Install the PCRE library.

    During testing we found that some systems don't have it.
    Note that we don't need to build it, we only need to make
    the source available.
    '''
    logger = context['logger']
    url = context['pcre']['url']
    srcpath = download_and_extract(context, url)
    context['pcre']['srcdir'] = srcpath


# ================================================================
# install_nginx
# ================================================================
def install_nginx(context):
    '''
    Install the nginx server.
    '''
    logger = context['logger']
    url = context['nginx']['url']
    srcpath = download_and_extract(context, url)

    # Configure
    rtfdir = context['dirs']['nginx']
    config_status = os.path.join(srcpath, 'objs', 'Makefile')
    if os.path.exists(config_status) is False:
        logger.info('configuring ' + srcpath)
        pcre_src = os.path.abspath(context['pcre']['srcdir'])
        user = context['username']
        sws = '--with-ipv6 --with-http_ssl_module --with-http_stub_status_module'
        sws += ' --with-pcre=%s' % (pcre_src)
        sws += ' --user=%s' % (context['username'])
        cmd = 'cd %s && ./configure --prefix="%s" --user="%s" %s' % (
            srcpath,
            rtfdir,
            user,
            sws,
        )
        runcmd(context, cmd)

    # Make
    semaphore = os.path.join(srcpath, 'objs', 'nginx')
    if os.path.exists(semaphore) is False:
        logger.info('running make in ' + srcpath)
        cmd = 'cd %s && make' % (srcpath)
        runcmd(context, cmd)

    # Make install
    semaphore = os.path.join(rtfdir, 'sbin', 'nginx')
    if os.path.exists(semaphore) is False:
        logger.info('running make install in ' + srcpath)
        cmd = 'cd %s && make install' % (srcpath)
        runcmd(context, cmd)

    # Conf file.
    confdir = os.path.join(rtfdir, 'conf')
    conf = os.path.join(confdir, 'nginx.conf')
    conf_orig = conf + '.orig'
    if os.path.exists(conf_orig) is False:
        logger.info('copy ' + conf)
        logger.info('  to ' + conf_orig)
        shutil.copyfile(conf, conf_orig)

    context['nginx']['conf'] = conf
    context['nginx']['confdir'] = confdir
    context['nginx']['ssl_cert_crt'] = os.path.join(confdir, context['sitename'] + '.crt')
    context['nginx']['ssl_cert_key'] = os.path.join(confdir, context['sitename'] + '.key')


# ================================================================
# install_uwsgi
# ================================================================
def install_uwsgi(context):
    '''
    Install the uwsgi server.
    '''
    # See if it has already been instdalled.
    uwsgi = os.path.join(context['dirs']['venv'], 'bin', 'uwsgi')
    if os.path.exists(uwsgi):
        return

    logger = context['logger']
    sts, out = pip_install(context, 'uwsgi', exit_on_error=False)

    if sts != 0:
        # The install failed. It may be the case that this is a
        # very old version of linux like CentOS 5.x. Try again
        # with a very old version uwsgi to see if that works.
        # If it doesn't fail.
        logger.info('install failed, will try a really old version')
        url = 'http://projects.unbit.it/downloads/uwsgi-1.2.6.tar.gz'
        context['uwsgi']['url'] = url
        
        srcpath = download_and_extract(context, url)
        cmd = '%s && cd %s && python uwsgiconfig.py --build' % (context['venv_activate_cmd'],
                                                                srcpath)
        runcmd(context, cmd)

        src = os.path.join(srcpath, 'uwsgi')
        logger.info('copy ' + src)
        logger.info('  to ' + uwsgi)
        shutil.copyfile(src, uwsgi)
        os.chmod(uwsgi, 0755)


# ================================================================
# install_django
# ================================================================
def install_django(context):
    '''
    Install and do basic configuration of django.
    '''
    logger = context['logger']
    pip_install(context, 'django')

    # Create the basic site.
    site_name = context['sitename']
    site_path = context['dirs']['django/project']
    if os.path.exists(site_path) is False:
        logger.info('starting django project "%s" in %s' % (site_name, site_path))
        cmd = ' %s && cd %s && django-admin.py startproject %s' % (context['venv_activate_cmd'],
                                                                   context['dirs']['django'],
                                                                   site_name)
        runcmd(context, cmd)

    for key in ['django/static', 'django/media', ]:
        path = context['dirs'][key]
        if os.path.exists(path) is False:
            makedirs(context, path)


# ================================================================
# install_libevent
# ================================================================
def install_libevent(context):
    '''
    Install the libevent library for use by memcached.
    '''
    logger = context['logger']
    url = context['libevent']['url']
    srcpath = download_and_extract(context, url)
    context['libevent']['srcdir'] = srcpath

    # Configure
    rtfdir = context['dirs']['rtf']
    config_status = os.path.join(srcpath, 'Makefile')
    if os.path.exists(config_status) is False:
        logger.info('configuring ' + srcpath)
        cmd = 'cd %s && ./configure --prefix="%s"' % (
            srcpath,
            rtfdir,
        )
        runcmd(context, cmd)

    # Make
    semaphore = os.path.join(srcpath, 'test', 'bench')
    if os.path.exists(semaphore) is False:
        logger.info('running make in ' + srcpath)
        cmd = 'cd %s && make' % (srcpath)
        runcmd(context, cmd)

    # Make install
    semaphore = os.path.join(rtfdir, 'lib', 'libevent.a')
    if os.path.exists(semaphore) is False:
        logger.info('running make install in ' + srcpath)
        cmd = 'cd %s && make install' % (srcpath)
        runcmd(context, cmd)


# ================================================================
# install_memcached
# ================================================================
def install_memcached(context):
    '''
    Install the memcached package.
    '''
    logger = context['logger']
    url = context['memcached']['url']
    srcpath = download_and_extract(context, url)
    context['memcached']['srcdir'] = srcpath

    # Configure
    rtfdir = context['dirs']['rtf']
    config_status = os.path.join(srcpath, 'Makefile')
    if os.path.exists(config_status) is False:
        logger.info('configuring ' + srcpath)
        cmd = 'cd %s && ./configure --prefix="%s"' % (
            srcpath,
            rtfdir,
        )
        runcmd(context, cmd)

    # Make
    semaphore = os.path.join(srcpath, 'memcached')
    if os.path.exists(semaphore) is False:
        logger.info('running make in ' + srcpath)
        cmd = 'cd %s && make' % (srcpath)
        runcmd(context, cmd)

    # Make install
    semaphore = os.path.join(rtfdir, 'bin', 'memcached')
    if os.path.exists(semaphore) is False:
        logger.info('running make install in ' + srcpath)
        cmd = 'cd %s && make install' % (srcpath)
        runcmd(context, cmd)

    # Install the python package.
    pkg = context['memcached']['pkg']
    pip_install(context, pkg)


# ================================================================
# fix_webserver_settings
# ================================================================
def fix_webserver_settings(context):
    '''
    Fix_Webserver_Settings dependencies between multiple packages.
    '''
    logger = context['logger']
    site_name = context['sitename']
    site_path = os.path.abspath(context['dirs']['django/project'])
    crt = context['nginx']['ssl_cert_crt']
    key = context['nginx']['ssl_cert_key']

    # The socket file use by the nginx/uwsgi stack.
    socket = os.path.join(site_path, site_name + '.sock')
    context['uwsgi']['socket'] = socket
    context['nginx']['socket'] = socket

    # The uwsgi configuration file.
    uwsgi_ini = os.path.join(site_path, 'uwsgi.ini')
    logfile = os.path.abspath(os.path.join(context['dirs']['log'], 'uwsgi.log'))
    pidfile = os.path.abspath(os.path.join(context['dirs']['log'], 'uwsgi.pid'))
    context['uwsgi']['logfile'] = logfile
    context['uwsgi']['pidfile'] = pidfile
    context['uwsgi']['inifile'] = uwsgi_ini
    if os.path.exists(uwsgi_ini) is False:
        logger.info('creating uwsgi ini file: ' + uwsgi_ini)
        ofp = open(uwsgi_ini, 'w')
        ofp. write('''[uwsgi]
# Django-related settings
# the base directory (full path)
chdir           = %s
# Django's wsgi file
module          = %s.wsgi:application
# the virtualenv (full path)
home            = %s

# process-related settings
# master
master          = true
# maximum number of worker processes
processes       = 10
# the socket (use the full path to be safe, used by nginx)
socket          = %s
# ... with appropriate permissions - may be needed
# chmod-socket    = 664
# clear environment on exit
vacuum          = true
# maximum requests
max-requests    = 5000

logto           = %s
logto2          = %s
pidfile         = %s
''' % (os.path.abspath(site_path),
       site_name,
       os.path.abspath(context['dirs']['venv']),
       context['uwsgi']['socket'],
       logfile,
       logfile,
       pidfile))
        ofp.close()

    # nginx setup
    nginx_html =  os.path.abspath(os.path.join(context['dirs']['nginx'], 'html'))
    uwsgi_params = os.path.join(site_path, 'uwsgi_params')
    uwsgi_params_src = os.path.join(context['dirs']['nginx'], 'conf', 'uwsgi_params')
    if os.path.exists(uwsgi_params) is False:
        logger.info('copy ' + uwsgi_params_src)
        logger.info('  to ' + uwsgi_params)
        shutil.copyfile(uwsgi_params_src, uwsgi_params)

    conf = context['nginx']['conf']
    tag = '# AUTOMATICALLY GENERATED'
    create = False
    if os.path.exists(conf) is True:
        ifp = open(conf, 'r')
        lines = ifp.readlines()
        ifp.close()
        create = len(lines) == 0 or lines[0].find(tag) < 0
    else:
        create = True

    logdir = os.path.abspath(context['dirs']['log'])
    pidfile = os.path.join(logdir, 'nginx.pid')
    context['nginx']['pidfile'] = pidfile
    if create:
        # Create the nginx conf file.
        logger.info('creating the nginx conf file: ' + conf)
        access_log = os.path.join(logdir, 'nginx_access.log')
        error_log =  os.path.join(logdir, 'nginx_error.log')
        user = context['username']
        group = grp.getgrgid(pwd.getpwnam(user).pw_gid).gr_name
        user_clause = 'user %s %s;' % (user, group)
        if context['login'] != 'root':
            user_clause = '#' + user_clause
        ofp = open(conf, 'w')
        ofp.write('''%s
worker_processes  1;
error_log  %s;
#error_log  %s debug;
pid %s;
%s

events {
    worker_connections  20;
}

http {
    include       mime.types;
    default_type  application/octet-stream;

    log_format  main  '$remote_addr - $remote_user [$time_local] "$request" '
                      '$status $body_bytes_sent "$http_referer" '
                      '"$http_user_agent" "$http_x_forwarded_for"';

    access_log %s  main;
    sendfile        on;
    keepalive_timeout  65;
    gzip on;

    # the upstream component nginx needs to connect to
    upstream django {
        server unix://%s;
    }

    # configuration of the HTTP server
    server {
        listen      %d;
        server_name %s;
        charset     utf-8;
        client_max_body_size 75M;

        # Django media
        location /media  {
            alias %s;
            # autoindex on; # useful for debugging
        }
    
        location /static {
            alias %s;
            # autoindex on; # useful for debugging
        }
    
        # Special debug
        location /nginx {
            alias %s;
            index  index.html index.htm;
        }

        # Finally, send all non-media requests to the Django server.
        location / {
            uwsgi_pass  django;
            include     %s;
        }
    }

    # configuration of the HTTPS server
    server {
        listen      %d;
        server_name %s;
        charset     utf-8;
        client_max_body_size 75M;

        ssl on;
        ssl_certificate %s;
        ssl_certificate_key %s;

        # Django media
        location /media  {
            alias %s;
        }
    
        location /static {
            alias %s;
        }
    
        # Special debug
        location /nginx {
            alias %s;
            index  index.html index.htm;
        }

        # Finally, send all non-media requests to the Django server.
        location / {
            uwsgi_pass  django;
            include     %s;
        }
    }
}
''' % (tag,
       error_log,
       error_log,
       pidfile,
       user_clause,
       access_log,
       context['nginx']['socket'],

       context['ports']['http'],
       context['hostname'],
       os.path.abspath(context['dirs']['django/media']),
       os.path.abspath(context['dirs']['django/static']),
       nginx_html,
       os.path.abspath(uwsgi_params),

       context['ports']['https'],
       context['hostname'],
       os.path.abspath(crt),
       os.path.abspath(key),
       os.path.abspath(context['dirs']['django/media']),
       os.path.abspath(context['dirs']['django/static']),
       nginx_html,
       os.path.abspath(uwsgi_params),
 ))

        ofp.close()

    # Create the certificates.
    if os.path.exists(crt) is False:
        logger.info('creating the self signed SSL certificate: ' + os.path.relpath(crt))
        cmd = 'openssl req -new -newkey rsa:4096 -days 365 -nodes -x509'
        cmd += ' -subj "/C=US/ST=Washington/L=Seattle/O=%s/OU=%s/CN=localhost"' % (context['sitename'],
                                                                                   context['appname'],)
        cmd += ' -keyout %s' % (os.path.abspath(key))
        cmd += ' -out %s' % (os.path.abspath(crt))
        runcmd(context, cmd)


# ================================================================
# read_file
# ================================================================
def read_file(ifn):
    '''
    Read a file.
    '''
    ifp = open(ifn, 'r')
    lines = ifp.readlines()
    ifp.close()
    return lines


# ================================================================
# write_file
# ================================================================
def write_file(ifn, lines):
    '''
    Write a file.
    '''
    ofp = open(ifn, 'w')
    ofp.write(''.join(lines))
    ofp.close()


# ================================================================
# configure_django_app
# ================================================================
def configure_django_app(context):
    '''
    Configure the django app.
    '''
    logger = context['logger']
    site_name = context['sitename']
    app_path = os.path.abspath(context['dirs']['django/app'])
    site_path = os.path.abspath(context['dirs']['django/project'])
    project_path = os.path.join(site_path, site_name)
    settings = os.path.join(project_path, 'settings.py')

    # Create the app skeleton.
    if os.path.exists(app_path) is False:
        logger.info('creating the django app: ' + context['appname'])
        cmd = '%s && cd %s && python manage.py startapp %s' % (context['venv_activate_cmd'],
                                                               site_path,
                                                               context['appname'])
        runcmd(context, cmd)
        
    # Add the app to the INSTALLED_APPS list in settings.py.
    lines = read_file(settings)
    for i in range(len(lines)):
        line = lines[i]
        if line.find('INSTALLED_APPS') == 0:
            beg = i
            appstr = "'" + context['appname'] + "'"
            write = True
            for j in range(i+1, len(lines)):
                line = lines[j]
                if line.find(')') == 0:
                    end = j
                    break
                elif line.find(appstr) > 0:
                    write = False
            if write is True:
                logger.info('adding app "%s" to INSTALL_APPS' % (context['appname']))
                stmts = ["    '%s',\n" % (context['appname']), ]
                newlines = lines[:end] + stmts + lines[end:]
                write_file(settings, newlines)
            break

    # Update the site URL to point to the app urls.
    project_urls = os.path.join(project_path, 'urls.py')
    lines = read_file(project_urls)
    for i in range(len(lines)):
        line = lines[i]
        if line.find('urlpatterns') == 0:
            beg = i
            app_url = context['appname'] + '.urls'
            write = True
            for j in range(i+1, len(lines)):
                line = lines[j]
                if line.find(')') == 0:
                    end = j
                    break
                elif line.find(app_url) > 0:
                    write = False
            if write is True:
                logger.info('adding app url reference "%s" to "%s"' % (app_url, project_urls))
                stmts = [
                    "    url('^$', include('%s')),\n" % (app_url),
                    "    url('^%s$', include('%s')),\n" % (context['appname'], app_url),
                ]
                newlines = lines[:end] + stmts + lines[end:]

                # Now add the extra code to make /static and /media visible.
                text = '''
# Allow /static directory access by django runserver for development.
# To enable this in production you must modify the nginx.conf file
# /static location by adding "autoindex on;".
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
urlpatterns += staticfiles_urlpatterns()

# Allow /media to be accessed by django runserver for development.
# To enable this in production you must modify the nginx.conf file
# /media location by adding "autoindex on;".
from django.conf import settings
import sys
import os
if 'runserver' in sys.argv or 'runserver_plus':
    urlpatterns += patterns('', 
                            url(r'^media/(.*)$',
                                'django.views.static.serve',
                                kwargs={'document_root': os.path.join(settings.BASE_DIR, 'media'),
                                        'show_indexes': True,}),
                            )
'''
                stmts = [line + '\n' for line in text.split('\n')]
                newlines += stmts
                write_file(project_urls, newlines)
            break

    # Create the app urls.
    app_urls = os.path.join(app_path, 'urls.py')
    if os.path.exists(app_urls) is False:
        logger.info('creating app index view url reference in "%s"' % (app_urls))
        ofp = open(app_urls, 'w')
        ofp.write('''from django.conf.urls import patterns, url
from %s import views

urlpatterns = patterns('',
    url(r'^$', views.index, name='index')
)
''' % (context['appname']))
        ofp.close()

    # Add the static path files: app.css and app.js.
    static_path = os.path.join(app_path, 'static', context['appname'])
    static_js_path = os.path.join(static_path, 'js')
    static_js = os.path.join(static_js_path, 'app.js')
    if os.path.exists(static_js) is False:
        logger.info('creating app custom js file: ' + os.path.relpath(static_js))
        makedirs(context, static_js_path)
        ofp = open(static_js, 'w')
        ofp.write('''// app custom javascript
$(function() {
    // Initialization code here.
});
''')
        ofp.close()

    static_css_path = os.path.join(static_path, 'css')
    static_css = os.path.join(static_css_path, 'app.css')
    if os.path.exists(static_css) is False:
        logger.info('creating app custom css file: ' + os.path.relpath(static_css))
        makedirs(context, static_css_path)
        ofp = open(static_css, 'w')
        ofp.write('/* app custom css */\n')
        ofp.close()

    # Add the <app_path>/templates/<app_name>/base__.html file.
    # It is the "extends" template for the app.
    base_path = os.path.join(app_path, 'templates', context['appname'])
    base_html = os.path.join(base_path, 'base__.html')
    if os.path.exists(base_html) is False:
        logger.info('creating the app base__.html file: %s' % (os.path.relpath(base_html)))
        makedirs(context, base_path)
        ofp = open(base_html, 'w')
        ofp.write('''{%% comment %%}
CITATION: https://oncampus.oberlin.edu/webteam/2012/09/architecture-django-templates
The base extendable template for pages.
{%% endcomment %%}
{%% load staticfiles i18n %%}
<!DOCTYPE html>
<html>
  <head>
    {%% block meta_tags %%}
      <meta http-equiv="cache-control" content="max-age=0">
      <meta http-equiv="cache-control" content="no-cache">
      <meta name="viewport" content="width=device-width, initial-scale=1.0">
    {%% endblock %%}

    <title>
      {%% block title %%}{%% trans "%s" %%}{%% endblock %%}
    </title>

    {%% block stylesheets %%}
      <link rel="stylesheet" type="text/css" href="{%% static "pkg/jquery-ui-themes-1.10.3/jquery-ui-themes-1.10.3/themes/ui-lightness/jquery.ui.theme.css" %%}">
      <link rel="stylesheet" type="text/css" href="{%% static "pkg/jquery-ui-themes-1.10.3/jquery-ui-themes-1.10.3/themes/ui-lightness/jquery-ui.min.css" %%}">
      <link rel="stylesheet" type="text/css" href="{%% static "pkg/tooltipster-master/tooltipster-master/css/tooltipster.css" %%}">
      <link rel="stylesheet" type="text/css" href="{%% static "pkg/DataTables-1.9.4/DataTables-1.9.4/media/css/jquery.dataTables.css" %%}">
      <link rel="stylesheet" type="text/css" href="{%% static "pkg/DataTables-1.9.4/DataTables-1.9.4/media/css/jquery.dataTables_themeroller.css" %%}">
      <link rel="stylesheet" type="text/css" href="{%% static "pkg/jquery.dynatree-1.2.5-all/dist/skin/ui.dynatree.css"%%}">
      <link rel="stylesheet" type="text/css" href="{%% static "pkg/bootstrap-3.0.3-dist/dist/css/bootstrap.min.css" %%}">
      <link rel="stylesheet" type="text/css" href="{%% static "pkg/bootstrap-3.0.3-dist/dist/css/bootstrap-theme.min.css" %%}">
      <link rel="stylesheet" type="text/css" href="{%% static "%s/css/app.css" %%}">
    {%% endblock %%}

    {%% block javascript %%}
      <script type="text/javascript" src="{%% static "js/jquery-2.0.3.min.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/jquery-ui-1.10.3/jquery-ui-1.10.3/ui/minified/jquery-ui.min.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/tooltipster-master/tooltipster-master/js/jquery.tooltipster.min.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/DataTables-1.9.4/DataTables-1.9.4/media/js/jquery.dataTables.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/jquery.dynatree-1.2.5-all/dist/jquery.dynatree.min.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/flot-0.8.2/flot/jquery.flot.min.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/flot-0.8.2/flot/jquery.flot.pie.min.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/flot-0.8.2/flot/jquery.flot.resize.min.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/flot-0.8.2/flot/jquery.flot.time.min.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/bootstrap-3.0.3-dist/dist/js/bootstrap.min.js" %%}"></script> 
      <script type="text/javascript" src="{%% static "pkg/CryptoJS-v3.1.2/rollups/aes.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/CryptoJS-v3.1.2/rollups/md5.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/CryptoJS-v3.1.2/rollups/sha512.js" %%}"></script>
      <script type="text/javascript" src="{%% static "pkg/CryptoJS-v3.1.2/rollups/tripledes.js" %%}"></script>
      <script type="text/javascript" src="{%% static "%s/js/app.js" %%}"></script>
    {%% endblock %%}

    {%% block extra_head %%}{%% endblock %%}
   </head>
   <body>
     <header class="navbar navbar-fixed-top">
       {%% block header %%}
       {%% endblock %%}
     </header>
     <div id="main" role="main">
       <div class="container">
         {%% block content %%}
         {%% endblock %%}
       </div>
     </div>
     <footer class="footer">
       {%% block footer %%}
       {%% endblock %%}
     </footer>
    </body>
</html>
''' % (context['appname'],
       context['appname'],
       context['appname'],
   ))
        ofp.close()

    # Add the <app_path>/templates/<app_name>/index.html file.
    # It is referenced in the view. It extends the base__.html
    # template.
    index_path = os.path.join(app_path, 'templates', context['appname'])
    index_html = os.path.join(index_path, 'index.html')
    if os.path.exists(index_html) is False:
        logger.info('creating the app index.html file: %s' % (os.path.relpath(index_html)))
        pkgs = geturls(context)
        fmt = '%s<tr><td align="left">%s</td><td align="left"><a href="%s">%s<a></td></tr>'
        pkgstr = '\n'.join([fmt % (10*' ',key, pkgs[key], pkgs[key]) for key in sorted(pkgs, key=str.lower)])

        # Directory tree for the django app.
        dirtree1 = []
        dirtree1.append('      <ul>')
        dirtree1 += walkdirs(os.path.relpath(site_path))
        dirtree1.append('      </ul>')
        assert len(dirtree1) > 2

        # Directory tree for rtf.
        dirtree2 = []
        dirtree2.append('      <ul>')
        dirtree2 += walkdirs(os.path.relpath(context['dirs']['rtf']))
        dirtree2.append('      </ul>')
        assert len(dirtree2) > 2

        makedirs(context, index_path)
        ofp = open(index_html, 'w')
        ofp.write('''{%% extends "%s/base__.html" %%}
{%% block content %%}
    <div style="margin: 10px;">
      <h3>Welcome!</h3>
      <p>
        Welcome to the default page of your %s application.
      </p>
      <p>
        It was created by <span tooltip="This demonstrates an interactive tooltipster tool tip.<br />Click on <a href='http://joelinoff.com'>http://joelinoff.com</a> to goto the download website.">mkwebsite-%s</span>.
        If you hover your mouse over mkwebsite-%s in the
        previous sentence an interactive tooltip (using the
        tooltipster package) will appear with a link to the download
        website.
      </p>
      <p>
        This page does not do anything useful yet but it provides the
        infrastructure to do all sorts of things without having to do
        any additional configuration.  It is tied into django so you
        can modify the contents using template parameters from the
        views which can, in turn, tie into the database. In addition
        it includes a number of static packages like jquery and
        jquery-ui so you can create complex client side behavior. It
        also includes the tooltipster package creating customized
        tooltips withoutm the title attribute, the dataTables package
        for creating amazing tables, the dynatree package for creating
        dynamic, interactive tree structures, the flot package for
        creating great charts, the bootstrap package for organizing
        information and the cryptojs package for doing client side
        encryption/decryption/hashing.
      </p>
      <p>
        With this infrastructure you can focus on writing attractive,
        interactive websites without worrying about the related
        administrative tasks.
      </p>

      <h3>Initial Application (%s)</h3>
      <p>
        Your initial django application which displays this page is
        installed in
        <code>%s</code>.
      </p>
      <p>
        If you right click on this page in your browser and display
        the source code you will see the client side tools defined for
        the project. They are a subset of what is available. Check the
        individual packages for details about additional capabilities.
      </p>
      <p>
        The directory tree below shows the files that were installed in your django project.
        It is displayed using the dynatree package to allow dynamic
        expansion and collapse of the directories.
      </p>
      <div class="dirtree">
       <code>%s</code>
%s
      </div>

      <h3>Release to Field (RTF)</h3>
      <p>
        The rtf area contains the server packages that were installed and
        the source of the static data. The django collectstatic function
        collects the static data from this area.
      <p>
      </p>
        The rtf directory tree from the time of the installation is shown below
        (also using the dynatree package).
      </p>
      <div class="dirtree">
       <code>%s</code>
%s
      </div>

      <h3>Packages</h3>
      <p>
        The table below (formated by datatables) shows the packages
        the all of the packages that mkwebsite installed for
        you.
      </p>
        <table id="pkg_list">
          <thead>
            <tr>
              <th tooltip="Abbreviated package name.">Package</th>
              <th tooltip="Server packages are installed on the server side, client packages operate on the client side (in the browser).">Type</th>
              <th>Version</th>
              <th>Project URL</th>
              <th>Brief Description</th>
            </tr>
          </thead>
          <tbody>
            <tr>
              <td align="left">bootstrap</td>
              <td align="right">3.0.3</td>
              <td align="left">client</td>
              <td align="left"><a href="http://getbootstrap.com">http://getbootstrap.com</a></td>
              <td align="left">sleek front end framework web development</td>
            </tr>
            
            <tr>
              <td align="left">cryptojs</td>
              <td align="right">3.1.2</td>
              <td align="left">client</td>
              <td align="left"><a href="http://code.google.com/p/crypto-js">http://code.google.com/p/crypto-js</a></td>
              <td align="left">cryptographic algorithms (like tripleDES) implemented in javascript</td>
            </tr>
            
            <tr>
              <td align="left">datatables</td>
              <td align="right">1.9.4</td>
              <td align="left">client</td>
              <td align="left"><a href="http://datatables.net">http://datatables.net</a></td>
              <td align="left">data tables plug in for jquery</td>
            </tr>
            
            <tr>
              <td align="left">django</td>
              <td align="right">1.6.1</td>
              <td align="left">server</td>
              <td align="left"><a href="https://www.djangoproject.com">https://www.djangoproject.com</a></td>
              <td align="left">python based server side web dev framework</td>
            </tr>
            
            <tr>
              <td align="left">dynatree</td>
              <td align="right">1.2.5</td>
              <td align="left">client</td>
              <td align="left"><a href="http://code.google.com/p/dynatree">http://code.google.com/p/dynatree</a></td>
              <td align="left">dynamic tree view plug in for jquery</td>
            </tr>
            
            <tr>
              <td align="left">flot</td>
              <td align="right">0.8.2</td>
              <td align="left">client</td>
              <td align="left"><a href="http://www.flotcharts.org">http://www.flotcharts.org</a></td>
              <td align="left">plotting/charting plug in for jquery</td>
            </tr>
            
            <tr>
              <td align="left">jquery</td>
              <td align="right">2.0.3</td>
              <td align="left">client</td>
              <td align="left"><a href="http://jquery.com">http://jquery.com</a></td>
              <td align="left">very useful javascript library for cross platform development</td>
            </tr>
            
            <tr>
              <td align="left">jquery-ui</td>
              <td align="right">1.10.3</td>
              <td align="left">client</td>
              <td align="left"><a href="http://jqueryui.com">http://jqueryui.com</a></td>
              <td align="left">user interface plug in for jquery</td>
            </tr>
            
            <tr>
              <td align="left">libevent</td>
              <td align="right">2.0.21</td>
              <td align="left">server</td>
              <td align="left"><a href="http://libevent.org">http://libevent.org</a></td>
              <td align="left">event notification library needed by memcached</td>
            </tr>
            
            <tr>
              <td align="left">memcached</td>
              <td align="right">1.4.17</td>
              <td align="left">server</td>
              <td align="left"><a href="http://memcached.org">http://memcached.org</a></td>
              <td align="left">high performance, distributed memory object caching system</td>
            </tr>
            
            <tr>
              <td align="left">nginx</td>
              <td align="right">1.4.4</td>
              <td align="left">server</td>
              <td align="left"><a href="http://nginx.org/en">http://nginx.org/en</a></td>
              <td align="left">high performance, lightweight web server</td>
            </tr>
            
            <tr>
              <td align="left">pcre</td>
              <td align="right">8.34</td>
              <td align="left">server</td>
              <td align="left"><a href="http://www.pcre.org">http://www.pcre.org</a></td>
              <td align="left">perl compatible regular expressions library needed by nginx</td>
            </tr>
            
            <tr>
              <td align="left">postgresql</td>
              <td align="right">9.3.2</td>
              <td align="left">server</td>
              <td align="left"><a href="http://www.postgresql.org">http://www.postgresql.org</a></td>
              <td align="left">high performance database</td>
            </tr>
            
            <tr>
              <td align="left">tooltipster</td>
              <td align="right">n/a</td>
              <td align="left">client</td>
              <td align="left"><a href="http://iamceege.github.io/tooltipster">http://iamceege.github.io/tooltipster</a></td>
              <td align="left">sleek tooltip plug in for jquery</td>
            </tr>
            
            <tr>
              <td align="left">uwsgi</td>
              <td align="right">1.2.6</td>
              <td align="left">server</td>
              <td align="left"><a href="https://uwsgi-docs.readthedocs.org/en/latest">https://uwsgi-docs.readthedocs.org/en/latest</a></td>
              <td align="left">python plug: web server gateway for nginx <--> django</td>
            </tr>
          </tbody>
        </table>
      <p>&nbsp;</p>
      <p>Enjoy!</p>
    </div>

    <!-- In production this would go into your app js file. -->
    <script type="text/javascript">
      $(function() {
        $('#pkg_list').dataTable({'bSort': true, 'bFilter': true});
        $('#pkgs_installed').dataTable({'bSort': true, 'bFilter': true});
        $('.dirtree').dynatree({});

        // Allow the use of tooltip="HTML" instead of title.
        $('[tooltip]').each(function(i, elem) {
            $(elem).tooltipster({
                contentAsHTML: true,
                interactive: true,
                content: $(elem).attr('tooltip')
            });
        });
      });
    </script>
{%% endblock %%}
''' % (context['appname'],
       context['appname'],
       VERSION,
       VERSION,
       context['appname'],
       app_path,
       os.path.abspath(context['dirs']['django']),
       '\n'.join(dirtree1),
       context['dirs']['rtf'],
       '\n'.join(dirtree2),
   ))
        ofp.close()

    # Add the view.
    app_views = os.path.join(app_path, 'views.py')
    lines = read_file(app_views)
    write = True
    for i in range(len(lines)):
        line = lines[i]
        if line.find('def index') == 0:
            write = False
            break

    if write == True:
        logger.info('creating the %s index view' % (context['appname']))
        ofp = open(app_views, 'w')
        ofp.write('''from django.shortcuts import render

def index(request):
    context = {}
    rendered_page = render(request, '%s/index.html', context)
    return rendered_page
''' % (context['appname']))
        ofp.close()
                        

# ================================================================
# configure_django
# ================================================================
def configure_django(context):
    '''
    Configure django for the database and static files.
    '''
    logger = context['logger']
    site_name = context['sitename']
    site_path = os.path.abspath(context['dirs']['django/project'])
    project_path = os.path.join(site_path, site_name)

    # Update STATIC_ROOT if it has not already been set in settings.py.
    settings =  os.path.join(project_path, 'settings.py')
    lines = read_file(settings)
    matches = filter(lambda x: x.find('STATIC_ROOT') == 0, lines)
    if len(matches) == 0:
        logger.info('setting STATIC_ROOT in django settings file:')
        logger.info(settings)
        backup(context, settings)
        lines.append("STATIC_ROOT = '%s'\n" % (os.path.abspath(context['dirs']['django/static'])))
        lines.append("STATICFILES_DIRS = ('%s', )\n" % (os.path.abspath(context['dirs']['static'])))
        write_file(settings, lines)

    # Update ALLOWED_HOSTS in settings.py.
    lines = read_file(settings)
    matches = filter(lambda x: x.find('ALLOWED_HOSTS = []') == 0, lines)
    if len(matches) == 1:
        logger.info('setting ALLOWED_HOSTS in django settings file:')
        logger.info(settings)
        backup(context, settings)
        for i in range(len(lines)):
            line = lines[i]
            if line.find('ALLOWED_HOSTS') == 0:
                lines[i] = "ALLOWED_HOSTS = ['*']\n"
                break
        write_file(settings, lines)

    # Add CACHES to settings.py based on memcached.
    sock_file = os.path.abspath(os.path.join(context['dirs']['log'], 'memcached.sock'))
    pid_file = os.path.abspath(os.path.join(context['dirs']['log'], 'memcached.pid'))
    context['memcached']['socket'] = sock_file
    context['memcached']['pidfile'] = pid_file
    lines = read_file(settings)
    matches = filter(lambda x: x.find('CACHES = {') >= 0, lines)
    if len(matches) == 0:
        logger.info('setting CACHES in django settings file:')
        logger.info(settings)
        backup(context, settings)
        newlines = [
            '\n',
            'CACHES = {\n',
            "    'default': {\n",
            "        'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',\n",
            "        'LOCATION': 'unix:%s',\n" % (sock_file),
            '    }\n',
            '}\n',
            ]
        lines += newlines
        write_file(settings, lines)
        # start the daemon: memcached -d -s <socket_file> -p <pid_file>
        
    # Update the DATABASES entry in settings.py.
    dbname = context['sitename']
    lines = read_file(settings)
    matches = filter(lambda x: x.find("'django.db.backends.sqlite3'") > 0, lines)
    if len(matches) == 1:
        logger.info('setting DATABASES in django settings file:')
        logger.info(settings)
        backup(context, settings)
        for i in range(len(lines)):
            line = lines[i]
            if line.find('DATABASES') == 0:
                beg = i
                for j in range(i+1, len(lines)):
                    line = lines[j]
                    if line.find('}') == 0:
                        end = j
                        break
                # Delete the existing DB entry and replace it with our new one.
                newlines = lines[:beg+1]
                newlines.append("    'default': {\n")
                newlines.append("        'ENGINE': 'django.db.backends.postgresql_psycopg2',\n")
                newlines.append("        'NAME': '%s',\n" % (dbname))
                newlines.append("        'USER': '%s',\n" % (context['username']))
                newlines.append("        'PASSWORD': '',\n")  # no password for now
                newlines.append("        'HOST': '%s',\n" % (context['hostname']))
                newlines.append("        'PORT': %d,\n" % (context['ports']['db']))
                newlines.append("    }\n")
                newlines += lines[end:]
                lines = newlines
                break
        write_file(settings, lines)

    # Create the DB and connect it to Django.
    # 1. start the db server
    # 2. use psl -l to determine whether the db exists.
    # 3. if it doesn't then create it and then sync django to it.
    # 4. stop the dbserver
    cmd = context['db']['start']
    if context['login'] == 'root':
        cmd = "su -m %s -c '%s'" % (context['username'], cmd)
        os.chmod(context['dirs']['log'], 0777)  # not sure if i could chown here instead
    logger.info('starting the database')
    runcmd(context, cmd)
    time.sleep(0.5)  # give it a bit of time to start

    logger.info('checking to see whether the database has been created')
    psql = os.path.join(context['dirs']['db/bin'], 'psql')
    cmd = psql + ' -l -t'
    if context['login'] == 'root':
        cmd = "su -m %s -c '%s'" % (context['username'], cmd)
    sts, out = runcmd(context, cmd, exit_on_error=False)
    if sts == 0:
        created = False
        for line in out.split('\n'):
            flds = line.split()
            if len(flds) > 0 and flds[0] == dbname:
                created = True
                break
    else:
        logger.info('stopping the database')
        cmd = context['db']['stop']
        if context['login'] == 'root':
            cmd = "su -m %s -c '%s'" % (context['username'], cmd)
        runcmd(context, cmd)
        logger.error('database operation failed, is it already running? try "ps -e -f | grep postgres"')
        sys.exit(1)

    if created is False:
        logger.info('creating the database "%s"' % (dbname))
        createdb = os.path.join(context['dirs']['db/bin'], 'createdb')
        cmd = createdb + ' ' + dbname + ' "site database" -O ' + context['username']
        if context['login'] == 'root':
            cmd = "su -m %s -c '%s'" % (context['username'], cmd)
        runcmd(context, cmd)
        time.sleep(0.5)  # give it a bit of time to work

        # Sync the DB with Django
        logger.info('syncing database with django')
        cmd = '%s && cd %s && python manage.py syncdb --noinput' % (context['venv_activate_cmd'],
                                                                    site_path)
        runcmd(context, cmd)
        time.sleep(0.5)

    # Create the super user.
    # Username: <username>
    # Password: <username>
    ofn = os.path.join(project_path, 'make-user.py')
    if os.path.exists(ofn) is False:
        logger.info('creating make-user.py')
        ofp = open(ofn, 'w')
        ofp.write(r'''import os
import sys
import django

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings')
from django.contrib.auth.models import User

def main():
    assert len(sys.argv) == 4
    username = sys.argv[1]
    password = sys.argv[2]
    superuser = sys.argv[3]

    user = User(username='%s' % (username))
    user.set_password('%s' % (password))
    if superuser.lower() in ['true', 't']:
        user.is_superuser = True
        user.is_staff = True

    try:
        user.save()
    except django.db.utils.IntegrityError as exc:
        print('user update failed for ' + username)
        print(str(exc))

if __name__ == '__main__':
    main()
''')
        # How to list all super users.
        #    import os
        #    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings')
        #    from django.contrib.auth.models import User
        #    from django.db.models import Q
        #    print(str(User.objects.filter(Q(is_superuser=True)).distinct()))
        ofp.close()
        os.chmod(ofn, 0755)

        username = context['username']
        password = context['username']
        superuser = True
        logger.info('creating super user: %s (password=%s)' % (username, password))
        cmd = '%s && cd %s && python make-user.py "%s" "%s" %s' % (context['venv_activate_cmd'],
                                                                   project_path,
                                                                   username,
                                                                   password,
                                                                   str(superuser))
        runcmd(context, cmd)

    logger.info('stopping the database')
    cmd = context['db']['stop']
    if context['login'] == 'root':
        cmd = "su -m %s -c '%s'" % (context['username'], cmd)
    runcmd(context, cmd)

    # Create the app.
    configure_django_app(context)

    # Collect static files.
    # This must be done after the app is created.
    # test/venv/lib/python2.7/site-packages/django/contrib/admin/static/admin/css/base.css
    logger.info('collecting static files for django')
    cmd = '%s && cd %s && python manage.py collectstatic --noinput' % (context['venv_activate_cmd'],
                                                                       site_path)
    runcmd(context, cmd)


# ================================================================
# install_mgr
# ================================================================
def install_mgr(context):
    '''
    Install the custom control script.

    The control script, like the installation script is written in
    python for portability.
    '''
    logger = context['logger']
    bindir = context['dirs']['bin']
    mgr = os.path.join(bindir, context['sitename'] + '-mgr.py')
    context['mgr'] = mgr

    if os.path.exists(mgr) is False:
        logger.info('install mgr ' + mgr)
        ofp = open(mgr, 'w')

        # Header
        ofp.write("""#!%s
'''
Automatically generated script that is used to control the website.
'''
import shutil
import socket
import subprocess
import sys
import os

# Global parameters.
DB_BIN_DIR = '%s'
DB_DATA = '%s'
DB_LOG = '%s'
DB_USER = '%s'
DB_SU = %s
DJANGO_PROJECT_DIR = '%s'
MEMCACHED_PID = '%s'
MEMCACHED_SOCK = '%s'
NGINX_BIN_DIR = '%s'
NGINX_PID = '%s'
PORT_DB = %d
PORT_HTTP = %d
PORT_HTTPS = %d
RTF_BIN_DIR = '%s'
VENV_DIR = '%s'
VENV_BIN_DIR = '%s'
UWSGI_INI = '%s'
UWSGI_LOG = '%s'
UWSGI_PID = '%s'
""" % (os.path.abspath(os.path.join(context['dirs']['venv/bin'], 'python')),
       context['dirs']['db/bin'],
       context['dirs']['dbdata'],
       context['db']['logfile'],
       context['username'],
       str(context['login'] == 'root'),
       os.path.abspath(context['dirs']['django/project']),
       os.path.abspath(context['memcached']['pidfile']),
       os.path.abspath(context['memcached']['socket']),
       context['dirs']['nginx/sbin'],
       context['nginx']['pidfile'],
       context['ports']['db'],
       context['ports']['http'],
       context['ports']['https'],
       context['dirs']['rtf/bin'],
       os.path.abspath(context['dirs']['venv']),
       os.path.abspath(context['dirs']['venv/bin']),
       context['uwsgi']['inifile'],
       os.path.abspath(context['uwsgi']['logfile']),
       context['uwsgi']['pidfile'],
       ))

        # The body -- this is fixed, it could be shipped as a separate
        # executable but that requires even more stuff for the user
        # keep track of.
        ofp.write(r"""

def init():
    '''
    Initialize the environment.

    After this is run all commands executed will operate with the new path
    settings and work as if they are in a virtual environment.
    '''
    path_dirs = [VENV_BIN_DIR, RTF_BIN_DIR, NGINX_BIN_DIR, DB_BIN_DIR,]
    for path_dir in path_dirs[::-1]:  # reverse
        os.environ['PATH'] = path_dir + os.pathsep + os.environ['PATH']

    # make sure that the VIRTUAL_ENV environment variable is set.
    os.environ['VIRTUAL_ENV'] = VENV_DIR


def port_is_open(address, port):
    '''
    Check to see if a port is open.
    '''
    try:
        socket.socket().connect((address, port))
        return True
    except socket.error:
        return False


def port_status(address='127.0.0.1', port=1):
    '''
    Return string 'open' or 'closed' for the port.
    '''
    return 'open' if port_is_open(address, port) is True else 'closed'


def run(title, cmd):
    '''
    Execute a shell command, return the output as a string as well as the status.
    '''
    if len(title) > 0:
        print(title)
    print('  cmd: ' + cmd)

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
    out, _ = process.communicate()
    sts = process.poll()

    out = out.rstrip()
    while len(out) > 1 and out[-1] == '\n':
        out = out[:-1]
    if len(out) > 0:
        for line in out.split('\n'):
            print('    %s' % (line.rstrip()))
    print('  exit_code = %d' % (sts))
    print('')


def hdr(title):
    '''
    print a header.
    '''
    print('# %s' % ('='*64))
    print('# %s' % (title))
    print('# %s' % ('='*64))


def collectstatic():
    '''
    Collect static files for django.
    '''
    hdr('collectstatic')
    run('collectstatic', 'cd %s && python manage.py collectstatic --noinput' % (DJANGO_PROJECT_DIR))


def syncdb():
    '''
    Sync the database for django.
    '''
    hdr('syncdb')
    run('syncb', 'cd %s && python manage.py syncdb --noinput' % (DJANGO_PROJECT_DIR))


def dbstatus():
    '''
    dbstatus command
    '''
    cmd = 'pg_ctl -D %s -l %s status' % (DB_DATA, DB_LOG)
    if DB_SU is True:
        cmd = "su -m %s -c '%s'" % (DB_USER, cmd)
    run('status db', cmd)


def dbstart():
    '''
    dbstart command
    '''
    cmd = 'pg_ctl -D %s -l %s start' % (DB_DATA, DB_LOG)
    if DB_SU is True:
        cmd = "su -m %s -c '%s'" % (DB_USER, cmd)
    run('start db', cmd)


def dbstop():
    '''
    dbstop command
    '''
    cmd = 'pg_ctl -D %s -l %s stop' % (DB_DATA, DB_LOG)
    if DB_SU is True:
        cmd = "su -m %s -c '%s'" % (DB_USER, cmd)
    run('stop db', cmd)


def memcached_start():
    '''
    start memcache
    '''
    cmd = 'memcached -d -s %s -P %s' % (MEMCACHED_SOCK, MEMCACHED_PID)
    if DB_SU is True:
        cmd = "su -m %s -c '%s'" % (DB_USER, cmd)
    run('start memcached', cmd)


def memcached_stop():
    '''
    stop memcache
    '''
    print('stop memcached')
    if os.path.exists(MEMCACHED_PID) is True:
        ifp = open(MEMCACHED_PID)
        data = ifp.read().strip()
        ifp.close()
        run('', 'kill -9 %s' % (data))
        if os.path.exists(MEMCACHED_PID) is True:
            os.remove(MEMCACHED_PID)
    else:
        print('  WARNING! PID file not found: %s\n' % (MEMCACHED_PID))


def uwsgi_start():
    '''
    start uwsgi
    '''
    run('start uwsgi', 'uwsgi -d %s --ini %s' % (UWSGI_LOG, UWSGI_INI))


def uwsgi_stop():
    '''
    stop uwsgi
    '''
    print('stop uwsgi')
    if os.path.exists(UWSGI_PID) is True:
        ifp = open(UWSGI_PID)
        data = ifp.read().strip()
        ifp.close()
        run('', 'kill -INT %s' % (data))
    else:
        print('  WARNING! PID file not found: %s\n' % (UWSGI_PID))


def start():
    '''
    start command
    '''
    hdr('start')
    dbstart()
    uwsgi_start()
    run('start nginx', 'nginx')
    memcached_start()


def stop():
    '''
    stop command
    '''
    hdr('stop')

    memcached_stop()
    run('stop nginx', 'nginx -s quit')
    uwsgi_stop()
    dbstop()


def restart():
    '''
    restart command
    '''
    hdr('restart')
    stop()
    start()


def reload():
    '''
    reload command
    '''
    hdr('reload')
    cmd = 'pg_ctl -D %s -l %s reload' % (DB_DATA, DB_LOG)
    if DB_SU is True:
        cmd = "su -m %s -c '%s'" % (DB_USER, cmd)
    run('reload db', cmd)
    # no reload for uwsgi
    # no reload for nginx
    # no reload for memcached


def status():
    '''
    status command
    '''
    hdr('status')

    print('ports')
    print('  db    %5d %s' % (PORT_DB, port_status(port=PORT_DB)))
    print('  http  %5d %s' % (PORT_HTTP, port_status(port=PORT_HTTP)))
    print('  https %5d %s' % (PORT_HTTPS, port_status(port=PORT_HTTPS)))
    print('')

    dbstatus()

    print('status uwsgi')
    if os.path.exists(UWSGI_PID) is True:
        ifp = open(UWSGI_PID)
        data = ifp.read().strip()
        ifp.close()
        print('  running (pid=%s)\n' % (data))
    else:
        print('  not running\n')

    print('status nginx')
    if os.path.exists(NGINX_PID) is True:
        ifp = open(NGINX_PID)
        data = ifp.read().strip()
        ifp.close()
        print('  running (pid=%s)\n' % (data))
    else:
        print('  not running\n')

    print('status memcached')
    if os.path.exists(MEMCACHED_PID) is True:
        ifp = open(MEMCACHED_PID)
        data = ifp.read().strip()
        ifp.close()
        print('  running (pid=%s)\n' % (data))
    else:
        print('  not running\n')
    

def version():
    '''
    version command
    '''
    hdr('version')
    run('version db', 'pg_ctl --version')
    run('version uwsgi', 'uwsgi --version')
    run('version nginx', 'nginx -V')
    run('version memcached', 'nginx -h | head -1')


def help():
    '''
    Help.
    '''
    print('''
    COMMANDS
        help           This help.

        collectstatic  Collect static files for django.
        syncdb         Sync the database for django.

        dbstart        Start the DB server.
        dbstop         Stop the DB server.
        dbstatus       Status of the DB server.

        reload         Reload the DB.
        restart        Restart the uwsgi, nginx, memcached and DB servers.
        start          Start the uwsgi, nginx, memcached and DB servers.
        status         Status of the uwsgi, nginx, memcached and DB servers.
        stop           Stop the uwsgi, nginx, memcached and DB servers.
        test           Test the environment.

    PORTS
        db     %5d  %s
        http   %5d  %s
        https  %5d  %s
''' % (PORT_DB, port_status(port=PORT_DB),
       PORT_HTTP, port_status(port=PORT_HTTP),
       PORT_HTTPS, port_status(port=PORT_HTTPS)))


def test():
    '''
    Test the setup.
    '''
    init()
    hdr('test')
    run('find uwsgi', 'which uwsgi')
    run('find nginx', 'which nginx')
    run('find pg_ctl', 'which pg_ctl')
    run('check venv pkgs', 'pip freeze')
    run('verify path', 'echo $PATH')
    run('find nginx', 'which memcached')


def main():
    '''
    Main entry point.
    '''
    init()
    valid = [
        'collectstatic',
        'syncdb',
        'dbstart',
        'dbstatus',
        'dbstop',
        'help',
        'memcached_start',  # not documented
        'memcached_stop',  # not documented
        'reload',
        'restart',
        'start',
        'status',
        'stop',
        'test',
        'uwsgi_start',  # not documented
        'uwsgi_stop',  # not documented
        'version',
    ]
    for i in range(1,len(sys.argv)):
        arg = sys.argv[i]
        if arg in valid:
            fctname = arg.lower() + '()'
            try:
                eval(fctname)
            except NameError:
                print('ERROR: unrecognized function: "%s"' % (fctname))
                sys.exit(1)
        else:
            print('ERROR: unrecognized command: "%s"' % (arg))
            sys.exit(1)

if __name__ == '__main__':
    main()
""")
        ofp.close()
        os.chmod(mgr, 0755)


# ================================================================
# geturls
# ================================================================
def geturls(context):
    '''
    Get the urls that were installed.
    '''
    urls = {}
    def lookforurls(rec, parents=[]):
        '''
        Scan through the data structure looking for
        string entried that start with 'http'.
        '''
        if rec is None:
            return

        parents.append(rec)  # push
        if type(rec) == dict:
            for key in rec:
                lookforurls(rec[key], parents)
        elif type(rec) == list:
            for item in rec:
                lookforurls(item, parents)
        elif type(rec) == str:
            if rec.find('http') == 0:
                key = os.path.basename(rec)
                if len(parents) > 1 and type(parents[-2]) == list and len(parents[-2]) == 3:
                    key = parents[-2][2]  # alias in static_pkgs
                urls[key] = rec
        parents.pop()

    for key in context:
        lookforurls(context[key])

    return urls


# ================================================================
# epilogue
# ================================================================
def epilogue(context, ofp=sys.stdout):
    '''
    The epilogue message.
    '''
    mgr = os.path.relpath(context['mgr'])
    pkgs = geturls(context)
    pkglist = []
    i = 0
    for key in sorted(pkgs, key=str.lower):
        i += 1
        pkglist.append((i, key))
    maxlen = len(mgr) + 14
    ofp.write('''
Website Bootstrap v%s

The installation completed on %s.

The following %d packages were installated:

%s
''' % (
    VERSION,
    str(datetime.datetime.now()),
    len(pkgs),
    '\n'.join(['    %2d %s' % (item[0], item[1]) for item in pkglist])
))

    ofp.write('''
You can start the django test server using:

    $ # URL: http://%s:%d
    $ # URL: http://%s:%d/%s
    $ # URL: http://%s:%d/admin
    $ %s
    $ %s
    $ %s
    $ cd %s
    $ python manage.py runserver 0.0.0.0:%d
''' % (
    context['hostname'],
    context['ports']['http'],

    context['hostname'],
    context['ports']['http'],
    context['appname'],

    context['hostname'],
    context['ports']['http'],

    mgr + ' stop',
    mgr + ' dbstart',
    context['venv_activate_path'],
    os.path.relpath(os.path.join(context['dirs']['django']), context['sitename']),
    context['ports']['http'],
))

    ofp.write('''
Or using uwsgi as follows to verify the stack: web client <--> uswgi <--> django

    $ # URL: http://%s:%d
    $ # URL: http://%s:%d/%s
    $ # URL: http://%s:%d/admin
    $ %s
    $ %s
    $ %s
    $ cd %s
    $ uwsgi --http :%d --module website.wsgi
''' % (
    context['hostname'],
    context['ports']['http'],

    context['hostname'],
    context['ports']['http'],
    context['appname'],

    context['hostname'],
    context['ports']['http'],

    mgr + ' stop',
    mgr + ' dbstart',
    context['venv_activate_path'],
    os.path.relpath(os.path.join(context['dirs']['django']), context['sitename']),
    context['ports']['http'],
))

    ofp.write('''
Django admin login:

    Username:  %s
    Password:  %s

DB admin login:

    $ %s
    $ %s
    $ %s -U %s %s
''' % (
    context['username'],
    context['username'],

    mgr + ' stop',
    mgr + ' dbstart',
    os.path.relpath(os.path.join(context['dirs']['db/bin']), 'psql'),
    context['username'],
    context['sitename'],
))

    ofp.write('''
Used ports:

    Database  %5d
    HTTP      %5d
    HTTPS     %5d
''' % (
    context['ports']['db'],
    context['ports']['http'],
    context['ports']['https'],
))

    ofp.write('''
Here are the control commands that are available.

    $ %-*s  Help.

    $ %-*s  Collect static files for django.
    $ %-*s  Sync the database for django.

    $ %-*s  Start the DB server.
    $ %-*s  Status of the DB server.
    $ %-*s  Stop the DB server.

    $ %-*s  Reload the DB.
    $ %-*s  Restart the uwsgi, nginx, memcached and DB servers.
    $ %-*s  Start the uwsgi, nginx, memcached and DB servers.
    $ %-*s  Status of the uwsgi, nginx, memcached and DB servers.
    $ %-*s  Stop the uwsgi, nginx, memcached and DB servers.
    $ %-*s  Test the environment.
''' % (
    maxlen, mgr + ' help',

    maxlen, mgr + ' collectstatic',
    maxlen, mgr + ' syncdb',

    maxlen, mgr + ' dbstart',
    maxlen, mgr + ' dbstatus',
    maxlen, mgr + ' dbstop',

    maxlen, mgr + ' restart',
    maxlen, mgr + ' reload',
    maxlen, mgr + ' start',
    maxlen, mgr + ' status',
    maxlen, mgr + ' stop',
    maxlen, mgr + ' test',
))

    ofp.write('''
To start using your new website follow these steps:

    1. Start the all of the web services: nginx, uwsgi and postgresql
       using the management tool that was generated for your site.

       These services run locally from the website directory and will
       not conflict with system wide services that might be running as
       long as the ports are different.

       $ %s start

    2. Navigate to the website:

       $ firefox http://localhost:%d
       $ firefox http://localhost:%d/%s
       $ firefox https://localhost:%d
       $ firefox https://localhost:%d/%s

       Self signed certificates were automatically created for
       HTTPS:
          CRT: %s
          KEY: %s

    3. The django app is located in %s.
       The templates and static subdirectories are also located in
       that directory. You can cd there to edit the app contents
       including the urls.

    4. The django project manage.py tool is in %s.
       To use it you must set your virtualenv like this:

       $ %s start
       $ .  %s
       $ cd %s
       $ python manage.py -h
       $ python manage.py collectstatic --noinput
       $ python manage.py syncdb --noinput
       $ python manage.py startapp mynewapp  # create another app

    5. The django site settings.py and urls.py files are in
       %s.

    6. You can access the database using psql like this:

       $ %s dbstart
       $ %s -U %s %s

    7. You can install additional python packages like this:

       $ . %s
       (venv) $ pip install boto twisted
       (venv) $ deactivate

    8. This information is captured in: %s.
''' % (
    # 1
    mgr,

    # 2
    context['ports']['http'],
    context['ports']['http'],
    context['appname'],
    context['ports']['https'],
    context['ports']['https'],
    context['appname'],
    os.path.relpath(context['nginx']['ssl_cert_crt']),
    os.path.relpath(context['nginx']['ssl_cert_key']),

    # 3
    os.path.relpath(context['dirs']['django/app']),

    # 4
    os.path.relpath(context['dirs']['django/project']),
    mgr,
    os.path.relpath(os.path.join(context['dirs']['venv/bin'], 'activate')),
    os.path.relpath(context['dirs']['django/project']),

    # 5
    os.path.relpath(os.path.join(context['dirs']['django/project'], context['sitename'])),

    # 6
    mgr,
    os.path.relpath(os.path.join(context['dirs']['db/bin']), 'psql'),
    context['username'],
    context['sitename'],

    # 7
    os.path.relpath(os.path.join(context['dirs']['venv/bin'], 'activate')),

    # 8
    os.path.relpath(os.path.join(context['dirs']['log'], 'install.log')),
   ))

    ofp.write('\n')
    ofp.flush()


# ================================================================
# update_context
# ================================================================
def update_context(context):
    '''
    Initialize the directories that must exist.

    This cannot be done until after the root directory is specified
    in the options.
    '''
    logger = context['logger']

    # Create the top level directory.
    topdir = os.path.join(context['rootdir'], context['sitename'])
    context['path'] = topdir
    makedirs(context, topdir)

    # Fix the dirs entries by prepending the rootdir.
    for dirname in context['dirs']:
        if dirname not in context['explicit-dirs']:
            val = context['dirs'][dirname]
            if val is not None:
                context['dirs'][dirname] = os.path.join(topdir, val)

    # Reset the django project directory with the proper sitename.
    project = os.path.join(context['dirs']['django'], context['sitename'])
    context['dirs']['django/project'] = project
    context['dirs']['django/media'] = os.path.join(project, 'media')
    context['dirs']['django/static'] = os.path.join(project, 'static')
    context['dirs']['django/app'] = os.path.join(project, context['appname'])

    # Define the absolute paths.
    for abspath in context['abspaths']:
        context['dirs'][abspath] = os.path.abspath(context['dirs'][abspath])

    # Create the virtual environment.
    venvdir = context['dirs']['venv']
    if os.path.exists(venvdir) is False:
        runcmd(context, 'virtualenv ' + venvdir)

    vap = os.path.join(venvdir, 'bin', 'activate')
    context['venv_activate_path'] = vap
    context['venv_activate_cmd'] = '. ' + os.path.abspath(vap)

    # Now find the site-packages directory and add it to the context.
    for root, dirs, files in os.walk(venvdir):
        for dirname in dirs:
            if dirname == 'site-packages':
                path = os.path.join(root, dirname)
                context['venv']['site-packages'] = os.path.abspath(path)
                break

    # Create the directories. This must done after all other processing.
    for key in sorted(context['dirs'], key=str.lower):
        if key not in context['nocreate']:
            dirpath = context['dirs'][key]
            makedirs(context, dirpath)

    # Add in the download names that are derived from the paths.
    for pkg in context['static_pkgs']:
        url = pkg[0]
        base = pkg[2] if len(pkg) > 2 else os.path.basename(url)
        path = os.path.join(context['dirs']['download'], base)
        pkg[1] = fixfn(path)

    # Get the UID, GID
    context['uid'] = pwd.getpwnam(context['username']).pw_uid
    context['gid'] = pwd.getpwnam(context['username']).pw_gid

    # Port permissions check.
    if context['login'] != 'root':
        errs = 0
        for key in sorted(context['ports'], key=str.lower):
            port = context['ports'][key]
            if port <= 1024:
                logger.error('invalid port (%d) specified for %s' % (port, key))
                errs += 1
        if errs > 0:
            logger.error('You must run as root and specify the --username switch to use a port in the [1..1024] range.')
            sys.exit(1)

# ================================================================
# checkenv
# ================================================================
def checkenv(context):
    '''
    Check the environment using heuristics to make sure that the
    infrastructure is setup correctly.
    '''
    logger = context['logger']
    checks = [
        ('pip installation check', 'pip -h'), 
        ('virtualenv installation check', 'virtualenv -h'), 
        ('make installation check', 'make -h'), 
        ('gcc installation check', 'gcc --help'), 
        ('openssl installation check', 'openssl -h'), 
    ]
    errs = 0
    for check in checks:
        sts, out = _runcmd(check[1])
        if sts != 0:
            logger.error('%s failed: %s' % (check[0], check[1]))
            errs += 1
        else:
            logger.info('%s succeeded: %s' % (check[0], check[1]))

    # port checks
    for key in sorted(context['ports'], key=str.lower):
        port = context['ports'][key]
        if port_is_open(context['hostname'], port):
            logger.error('%s port %d is open - please close it' % (key, port))
            errs += 1

    if errs > 0:
        logger.error('%d environment errors: please fix them before proceeding' % (errs))
        sys.exit(1)

    if context['sitename'] == context['appname']:
        logger.error('the website name cannot be the same as the appname: ' + context['sitename'])
        sys.exit(1)


# ================================================================
# init_context
# ================================================================
def init_context(logger):
    '''
    Initialize the context. The context is the information
    used to generate the website. It includes directory names,
    a reference to the logger, the opts and anything else that
    might be useful.
    '''
    username = getpass.getuser()
    context = {
        'opts': None,
        'logger': logger,
        'rootdir': '/opt/website',
        'sitename': 'website',
        'path': None,  # composed of rootdir + sitename, must be created after opts processing
        'appname': 'myapp',
        'login': username,  # permanent record
        'username': username,  # can by changed by --user
        'uid': pwd.getpwnam(username).pw_uid,  # associated with the username not the login
        'gid': pwd.getpwnam(username).pw_gid,  # ibid.
        'hostname': 'localhost',
        'explicit-dirs': [],

        'ports': {
            'http': 9110,
            'https': 9111,
            'db': 9112,
        },

        'db': {
            'url': 'http://ftp.postgresql.org/pub/source/v9.3.2/postgresql-9.3.2.tar.bz2',
            'logfile': None,
            'start': None,
            'stop': None,
        },

        'pcre': {
            'url': 'http://downloads.sourceforge.net/project/pcre/pcre/8.34/pcre-8.34.tar.bz2',
            'srcdir': None,
        },

        'libevent': {
            'url': 'https://github.com/downloads/libevent/libevent/libevent-2.0.21-stable.tar.gz',
            'srcdir': None,
        },

        'memcached': {
            'url': 'http://www.memcached.org/files/memcached-1.4.17.tar.gz',
            'pkg': 'python-memcached',  # python package
            'srcdir': None,
            'pidfile': None,
            'socket': None,
        },

        'nginx': {
            'url': 'http://nginx.org/download/nginx-1.4.4.tar.gz',
            'pidfile': None,
            'conf': None,
            'confdir': None,
            'socket': None,
            'ssl_cert_crt': None,
            'ssl_cert_key': None,
        },

        'uwsgi': {
            'logfile': None,
            'pidfile': None,
            'inifile': None,
            'socket': None,
            'url': None,  # modern versions are installed as part of the python virtualenv using pip
        },

        'venv': {
            'site-packages': None,
            'installed': [],
        },

        'abspaths': ['download', 'db', 'db/bin', 'dbdata', 'log', 'nginx', 'nginx/sbin', 'rtf', 'rtf/bin', 'rtf/lib', ],
        'nocreate': ['django/project', 'django/static', 'django/media', 'django/app',],

        'dirs': {
            'bin': os.path.join('rtf', 'bin'),
            'db': os.path.join('rtf', 'db'),
            'db/bin': os.path.join('rtf', 'db', 'bin'),
            'dbdata': os.path.join('rtf', 'db', 'dbdata'),
            'django': 'django',
            'django/app': None,
            'django/project': None,
            'django/media': None,
            'django/static': None,
            'download': 'download',
            'log': os.path.join('rtf', 'log'),
            'nginx': os.path.join('rtf', 'nginx'),
            'nginx/sbin': os.path.join('rtf', 'nginx', 'sbin'),
            'rtf': 'rtf',
            'rtf/bin': os.path.join('rtf', 'bin'),
            'rtf/lib': os.path.join('rtf', 'lib'),
            'src': 'src',
            'static': os.path.join('rtf', 'static'),
            'static/css': os.path.join('rtf', 'static', 'css'),
            'static/img': os.path.join('rtf', 'static', 'img'),
            'static/js': os.path.join('rtf', 'static', 'js'),
            'static/pkg': os.path.join('rtf', 'static', 'pkg'),
            'venv': 'venv',
            'venv/bin': os.path.join('venv', 'bin'),
            'venv/lib': os.path.join('venv', 'lib'),
        },

        'static_pkgs': [
            ['http://code.jquery.com/jquery-2.0.3.js', None, ],
            ['http://code.jquery.com/jquery-2.0.3.min.js', None, ],
            ['http://code.jquery.com/jquery-2.0.3.min.map', None, ],
            ['http://jqueryui.com/resources/download/jquery-ui-1.10.3.zip', None, ],
            ['http://jqueryui.com/resources/download/jquery-ui-themes-1.10.3.zip', None, ],
            ['https://github.com/iamceege/tooltipster/archive/master.zip', None, 'tooltipster-master.zip',],  # i prefer tooltipster for tooltips
            ['http://crypto-js.googlecode.com/files/CryptoJS%20v3.1.2.zip', None, ],
            ['http://datatables.net/releases/DataTables-1.9.4.zip', None, ],
            ['http://dynatree.googlecode.com/files/jquery.dynatree-1.2.5-all.zip', None, ],
            ['http://www.flotcharts.org/downloads/flot-0.8.2.zip', None, ],
            ['https://github.com/twbs/bootstrap/releases/download/v3.0.3/bootstrap-3.0.3-dist.zip', None, ],
            ]
        }

    return context


# ================================================================
# getopts
# ================================================================
def getopts(context):
    '''
    Get the command line options.
    '''
    # Setup the CLI args help to use the module help
    # and add an epilog.
    # This helps avoid redundant help.
    this = os.path.basename(sys.argv[0])
    description = ('description:%s' % '\n  '.join(__doc__.split('\n')))
    epilog = '''author:
    Joe Linoff'''
    rawd = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=rawd,
                                     description=description,
                                     epilog=epilog,
                                     prefix_chars='-+')

    parser.add_argument('-a', '--appname',
                        action='store',
                        default=context['appname'],
                        metavar=('DIR'),
                        help='appname directory in django, default="%(default)s"')

    parser.add_argument('-d', '--download-dir',
                        action='store',
                        default=context['dirs']['download'],
                        metavar=('DIR'),
                        help='specify an explicit downloads directory independent of the site, default="<website>/%(default)s"')

    parser.add_argument('--port-db',
                        action='store',
                        type=int,
                        default=context['ports']['db'],
                        metavar=('PORT'),
                        help='database port, default=%(default)s')

    parser.add_argument('--port-http',
                        action='store',
                        type=int,
                        default=context['ports']['http'],
                        metavar=('PORT'),
                        help='HTTP port, default=%(default)s')

    parser.add_argument('--port-https',
                        action='store',
                        type=int,
                        default=context['ports']['https'],
                        metavar=('PORT'),
                        help='HTTPS port, default=%(default)s')

    parser.add_argument('-r', '--root-dir',
                        action='store',
                        default=context['rootdir'],
                        metavar=('DIR'),
                        help='root directory, default="%(default)s"')

    parser.add_argument('-s', '--sitename',
                        action='store',
                        default=context['sitename'],
                        metavar=('DIR'),
                        help='sitename directory in django, default="%(default)s"')

    parser.add_argument('--start',
                        action='store_true',
                        help='start the web services after the installation')

    parser.add_argument('-u', '--username',
                        action='store',
                        default=context['username'],
                        metavar=('USER'),
                        help='non-privileged user for nginx and postgres, default="%(default)s"')

    parser.add_argument('-v', '--verbose',
                        action='count',
                        help='level of verbosity')

    # Display the version number and exit.
    parser.add_argument('-V', '--version',
                        action='version',
                        version='%(prog)s - v' + VERSION)

    opts = parser.parse_args()

    # Update the context.
    for port in sorted(context['ports'], key=str.lower):
        optval = eval('opts.port_' + port)  # derive the opts name from the port key
        if context['ports'][port] != optval:
            context['ports'][port] = optval

    if context['rootdir'] != opts.root_dir:
        context['rootdir'] = opts.root_dir

    if context['sitename'] != opts.sitename:
        context['sitename'] = opts.sitename

    if context['appname'] != opts.appname:
        context['appname'] = opts.appname

    if context['username'] != opts.username:
        context['username'] = opts.username

    if context['dirs']['download'] != opts.download_dir:
        context['dirs']['download'] = opts.download_dir
        context['explicit-dirs'].append('download')

    if opts.verbose:
        context['logger'].setLevel(logging.INFO)
    else:
        context['logger'].setLevel(logging.ERROR)

    return opts


# ================================================================
# main
# ================================================================
def main():
    '''
    Main entry point.
    '''
    start = datetime.datetime.now()
    logger = init_logger('install')
    context = init_context(logger)
    context['opts'] = getopts(context)
    logger.info('website bootstrap v' + VERSION)
    logger.info('creating website "%s"' % (context['sitename']))
    checkenv(context)

    update_context(context)
    logger.info('context initial state')
    if context['opts'].verbose > 1:
        pprint(context)
        sys.stdout.flush()

    # Some useful pip packages
    pip_install(context, 'yolk')
    pip_install(context, 'south')  # for schema migration
    pip_install(context, 'fabric')  # just plain useful, include paramiko

    install_static(context)
    install_db(context)
    install_pcre(context)
    install_nginx(context)
    install_uwsgi(context)
    install_django(context)

    # memcached support
    install_libevent(context)
    install_memcached(context)

    fix_webserver_settings(context)
    configure_django(context)

    logger.info('TODO: install examples')
    install_mgr(context)  # the management script

    logger.info('context final state')
    if context['opts'].verbose > 1:
        pprint(context)
        sys.stdout.flush()

    epilogue(context)

    install_log = os.path.join(context['dirs']['log'], 'install.log')
    logger.info('writing install info to ' + os.path.relpath(install_log))
    ofp = open(install_log, 'w')
    epilogue(context, ofp)
    ofp.close()
    os.chmod(install_log, 0644)

    if context['opts'].start is True:
        logger.info('starting the services: ' + os.path.relpath(context['mgr']))

        cmd = context['mgr'] + ' stop'
        runcmd(context, cmd)

        cmd = context['mgr'] + ' start'
        runcmd(context, cmd)

    elapsed = datetime.datetime.now() - start
    logger.info('elapsed time %.1f' % (_total_seconds(elapsed)))

    
if __name__ == '__main__':
    main()
